@@ -7,6 +7,7 @@ import (
 	"strings"
 
 	corev1 "k8s.io/api/core/v1"
+	apiextensionsclient "k8s.io/apiextensions-apiserver/pkg/client/clientset/clientset"
 	"k8s.io/apimachinery/pkg/api/errors"
 	"k8s.io/apimachinery/pkg/api/meta"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
@@ -14,7 +15,7 @@ import (
 	appsinformer "k8s.io/client-go/informers/apps/v1"
 	coreinformer "k8s.io/client-go/informers/core/v1"
 	"k8s.io/client-go/kubernetes"
-	"k8s.io/klog"
+	"k8s.io/klog/v2"
 
 	"github.com/openshift/library-go/pkg/controller/factory"
 	"github.com/openshift/library-go/pkg/operator/events"
@@ -40,6 +41,10 @@ const (
 )
 
 var (
+	crdStaticFiles = []string{
+		"manifests/klusterlet/0000_01_work.open-cluster-management.io_appliedmanifestworks.crd.yaml",
+	}
+
 	staticResourceFiles = []string{
 		"manifests/klusterlet/klusterlet-registration-serviceaccount.yaml",
 		"manifests/klusterlet/klusterlet-registration-clusterrole.yaml",
@@ -59,16 +64,18 @@ var (
 )
 
 type klusterletController struct {
-	klusterletClient  operatorv1client.KlusterletInterface
-	klusterletLister  operatorlister.KlusterletLister
-	kubeClient        kubernetes.Interface
-	kubeVersion       *version.Version
-	operatorNamespace string
+	klusterletClient   operatorv1client.KlusterletInterface
+	klusterletLister   operatorlister.KlusterletLister
+	kubeClient         kubernetes.Interface
+	apiExtensionClient apiextensionsclient.Interface
+	kubeVersion        *version.Version
+	operatorNamespace  string
 }
 
 // NewKlusterletController construct klusterlet controller
 func NewKlusterletController(
 	kubeClient kubernetes.Interface,
+	apiExtensionClient apiextensionsclient.Interface,
 	klusterletClient operatorv1client.KlusterletInterface,
 	klusterletInformer operatorinformer.KlusterletInformer,
 	secretInformer coreinformer.SecretInformer,
@@ -77,11 +84,12 @@ func NewKlusterletController(
 	operatorNamespace string,
 	recorder events.Recorder) factory.Controller {
 	controller := &klusterletController{
-		kubeClient:        kubeClient,
-		klusterletClient:  klusterletClient,
-		klusterletLister:  klusterletInformer.Lister(),
-		kubeVersion:       kubeVersion,
-		operatorNamespace: operatorNamespace,
+		kubeClient:         kubeClient,
+		apiExtensionClient: apiExtensionClient,
+		klusterletClient:   klusterletClient,
+		klusterletLister:   klusterletInformer.Lister(),
+		kubeVersion:        kubeVersion,
+		operatorNamespace:  operatorNamespace,
 	}
 
 	return factory.New().WithSync(controller.sync).
@@ -126,8 +134,8 @@ func (n *klusterletController) sync(ctx context.Context, controllerContext facto
 		RegistrationImage:         klusterlet.Spec.RegistrationImagePullSpec,
 		WorkImage:                 klusterlet.Spec.WorkImagePullSpec,
 		ClusterName:               klusterlet.Spec.ClusterName,
-		BootStrapKubeConfigSecret: helpers.BootstrapHubKubeConfigSecret,
-		HubKubeConfigSecret:       helpers.HubKubeConfigSecret,
+		BootStrapKubeConfigSecret: helpers.BootstrapHubKubeConfig,
+		HubKubeConfigSecret:       helpers.HubKubeConfig,
 		ExternalServerURL:         getServersFromKlusterlet(klusterlet),
 		OperatorNamespace:         n.operatorNamespace,
 	}
@@ -169,14 +177,14 @@ func (n *klusterletController) sync(ctx context.Context, controllerContext facto
 			ObjectMeta: metav1.ObjectMeta{Name: config.KlusterletNamespace},
 		}, metav1.CreateOptions{})
 		if createErr != nil {
-			helpers.UpdateKlusterletStatus(ctx, n.klusterletClient, klusterletName, helpers.UpdateKlusterletConditionFn(operatorapiv1.StatusCondition{
+			_, _, _ = helpers.UpdateKlusterletStatus(ctx, n.klusterletClient, klusterletName, helpers.UpdateKlusterletConditionFn(metav1.Condition{
 				Type: klusterletApplied, Status: metav1.ConditionFalse, Reason: "KlusterletApplyFailed",
 				Message: fmt.Sprintf("Failed to create namespace %q: %v", config.KlusterletNamespace, createErr),
 			}))
 			return createErr
 		}
 	case err != nil:
-		helpers.UpdateKlusterletStatus(ctx, n.klusterletClient, klusterletName, helpers.UpdateKlusterletConditionFn(operatorapiv1.StatusCondition{
+		_, _, _ = helpers.UpdateKlusterletStatus(ctx, n.klusterletClient, klusterletName, helpers.UpdateKlusterletConditionFn(metav1.Condition{
 			Type: klusterletApplied, Status: metav1.ConditionFalse, Reason: "KlusterletApplyFailed",
 			Message: fmt.Sprintf("Failed to get namespace %q: %v", config.KlusterletNamespace, err),
 		}))
@@ -195,7 +203,7 @@ func (n *klusterletController) sync(ctx context.Context, controllerContext facto
 	)
 
 	if err != nil {
-		helpers.UpdateKlusterletStatus(ctx, n.klusterletClient, klusterletName, helpers.UpdateKlusterletConditionFn(operatorapiv1.StatusCondition{
+		_, _, _ = helpers.UpdateKlusterletStatus(ctx, n.klusterletClient, klusterletName, helpers.UpdateKlusterletConditionFn(metav1.Condition{
 			Type: klusterletApplied, Status: metav1.ConditionFalse, Reason: "KlusterletApplyFailed",
 			Message: fmt.Sprintf("Failed to sync image pull secret to namespace %q: %v", config.KlusterletNamespace, err),
 		}))
@@ -223,13 +231,14 @@ func (n *klusterletController) sync(ctx context.Context, controllerContext facto
 	}
 
 	// Apply static files
+	appliedStaticFiles := append(crdStaticFiles, staticResourceFiles...)
 	resourceResults := resourceapply.ApplyDirectly(
-		resourceapply.NewKubeClientHolder(n.kubeClient),
+		resourceapply.NewKubeClientHolder(n.kubeClient).WithAPIExtensionsClient(n.apiExtensionClient),
 		controllerContext.Recorder(),
 		func(name string) ([]byte, error) {
 			return assets.MustCreateAssetFromTemplate(name, bindata.MustAsset(filepath.Join("", name)), config).Data, nil
 		},
-		staticResourceFiles...,
+		appliedStaticFiles...,
 	)
 
 	for _, result := range resourceResults {
@@ -240,7 +249,7 @@ func (n *klusterletController) sync(ctx context.Context, controllerContext facto
 
 	if len(errs) > 0 {
 		applyErrors := operatorhelpers.NewMultiLineAggregate(errs)
-		helpers.UpdateKlusterletStatus(ctx, n.klusterletClient, klusterletName, helpers.UpdateKlusterletConditionFn(operatorapiv1.StatusCondition{
+		_, _, _ = helpers.UpdateKlusterletStatus(ctx, n.klusterletClient, klusterletName, helpers.UpdateKlusterletConditionFn(metav1.Condition{
 			Type: klusterletApplied, Status: metav1.ConditionFalse, Reason: "KlusterletApplyFailed",
 			Message: applyErrors.Error(),
 		}))
@@ -248,27 +257,27 @@ func (n *klusterletController) sync(ctx context.Context, controllerContext facto
 	}
 
 	// Create hub config secret
-	hubSecret, err := n.kubeClient.CoreV1().Secrets(config.KlusterletNamespace).Get(ctx, helpers.HubKubeConfigSecret, metav1.GetOptions{})
+	hubSecret, err := n.kubeClient.CoreV1().Secrets(config.KlusterletNamespace).Get(ctx, helpers.HubKubeConfig, metav1.GetOptions{})
 	switch {
 	case errors.IsNotFound(err):
 		// Create an empty secret with placeholder
 		hubSecret = &corev1.Secret{
 			ObjectMeta: metav1.ObjectMeta{
-				Name:      helpers.HubKubeConfigSecret,
+				Name:      helpers.HubKubeConfig,
 				Namespace: config.KlusterletNamespace,
 			},
 			Data: map[string][]byte{"placeholder": []byte("placeholder")},
 		}
 		hubSecret, err = n.kubeClient.CoreV1().Secrets(config.KlusterletNamespace).Create(ctx, hubSecret, metav1.CreateOptions{})
 		if err != nil {
-			helpers.UpdateKlusterletStatus(ctx, n.klusterletClient, klusterletName, helpers.UpdateKlusterletConditionFn(operatorapiv1.StatusCondition{
+			_, _, _ = helpers.UpdateKlusterletStatus(ctx, n.klusterletClient, klusterletName, helpers.UpdateKlusterletConditionFn(metav1.Condition{
 				Type: klusterletApplied, Status: metav1.ConditionFalse, Reason: "KlusterletApplyFailed",
 				Message: fmt.Sprintf("Failed to create hub kubeconfig secret -n %q %q: %v", hubSecret.Namespace, hubSecret.Name, err),
 			}))
 			return err
 		}
 	case err != nil:
-		helpers.UpdateKlusterletStatus(ctx, n.klusterletClient, klusterletName, helpers.UpdateKlusterletConditionFn(operatorapiv1.StatusCondition{
+		_, _, _ = helpers.UpdateKlusterletStatus(ctx, n.klusterletClient, klusterletName, helpers.UpdateKlusterletConditionFn(metav1.Condition{
 			Type: klusterletApplied, Status: metav1.ConditionFalse, Reason: "KlusterletApplyFailed",
 			Message: fmt.Sprintf("Failed to get hub kubeconfig secret with error %v", err),
 		}))
@@ -285,7 +294,7 @@ func (n *klusterletController) sync(ctx context.Context, controllerContext facto
 		controllerContext.Recorder(),
 		"manifests/klusterlet/klusterlet-registration-deployment.yaml")
 	if err != nil {
-		helpers.UpdateKlusterletStatus(ctx, n.klusterletClient, klusterletName, helpers.UpdateKlusterletConditionFn(operatorapiv1.StatusCondition{
+		_, _, _ = helpers.UpdateKlusterletStatus(ctx, n.klusterletClient, klusterletName, helpers.UpdateKlusterletConditionFn(metav1.Condition{
 			Type: klusterletApplied, Status: metav1.ConditionFalse, Reason: "KlusterletApplyFailed",
 			Message: fmt.Sprintf("Failed to deploy registration deployment with error %v", err),
 		}))
@@ -310,7 +319,7 @@ func (n *klusterletController) sync(ctx context.Context, controllerContext facto
 		controllerContext.Recorder(),
 		"manifests/klusterlet/klusterlet-work-deployment.yaml")
 	if err != nil {
-		helpers.UpdateKlusterletStatus(ctx, n.klusterletClient, klusterletName, helpers.UpdateKlusterletConditionFn(operatorapiv1.StatusCondition{
+		_, _, _ = helpers.UpdateKlusterletStatus(ctx, n.klusterletClient, klusterletName, helpers.UpdateKlusterletConditionFn(metav1.Condition{
 			Type: klusterletApplied, Status: metav1.ConditionFalse, Reason: "KlusterletApplyFailed",
 			Message: fmt.Sprintf("Failed to deploy work deployment with error %v", err),
 		}))
@@ -319,8 +328,8 @@ func (n *klusterletController) sync(ctx context.Context, controllerContext facto
 	observedKlusterletGeneration := klusterlet.Generation
 
 	// if we get here, we have successfully applied everything and should indicate that
-	helpers.UpdateKlusterletStatus(ctx, n.klusterletClient, klusterletName,
-		helpers.UpdateKlusterletConditionFn(operatorapiv1.StatusCondition{
+	_, _, _ = helpers.UpdateKlusterletStatus(ctx, n.klusterletClient, klusterletName,
+		helpers.UpdateKlusterletConditionFn(metav1.Condition{
 			Type: klusterletApplied, Status: metav1.ConditionTrue, Reason: "KlusterletApplied",
 			Message: "Klusterlet Component Applied"}),
 		helpers.UpdateKlusterletGenerationsFn(registrationGeneration, workGeneration),
@@ -358,7 +367,7 @@ func (n *klusterletController) cleanUp(ctx context.Context, controllerContext fa
 		err := helpers.CleanUpStaticObject(
 			ctx,
 			n.kubeClient,
-			nil,
+			n.apiExtensionClient,
 			nil,
 			func(name string) ([]byte, error) {
 				return assets.MustCreateAssetFromTemplate(name, bindata.MustAsset(filepath.Join("", name)), config).Data, nil
@@ -390,7 +399,13 @@ func (n *klusterletController) cleanUp(ctx context.Context, controllerContext fa
 			}
 		}
 	}
-	return nil
+
+	// remove the klusterlet namespace
+	err = n.kubeClient.CoreV1().Namespaces().Delete(ctx, config.KlusterletNamespace, metav1.DeleteOptions{})
+	if errors.IsNotFound(err) {
+		return nil
+	}
+	return err
 }
 
 func (n *klusterletController) removeKlusterletFinalizer(ctx context.Context, deploy *operatorapiv1.Klusterlet) error {
